import unittest

from ..results import Results
from ..problem import Problem
from ..benchmark_functions import Ackley

try:
    from ..algorithm_pymoo import Pymoo

    from pymoo.algorithms.soo.nonconvex.ga import GA
    from pymoo.algorithms.moo.nsga2 import NSGA2
    from pymoo.algorithms.soo.nonconvex.isres import ISRES
    from pymoo.algorithms.soo.nonconvex.nelder import NelderMead
    from pymoo.algorithms.soo.nonconvex.de import DE

    from pymoo.operators.sampling.lhs import LHS

    from pymoo.operators.crossover.sbx import SimulatedBinaryCrossover
    from pymoo.operators.mutation.pm import PolynomialMutation
    from pymoo.operators.sampling.rnd import FloatRandomSampling

    __pymoo__ = True
except ImportError:
    print("pymoo is not present test skipped")
    __pymoo__ = False


class ProblemConstraint(Problem):
    def set(self):
        self.name = 'ProblemConstraint'

        self.parameters = [{'name': 'x', 'initial_value': 1.0, 'bounds': [-2, 2]},
                           {'name': 'y', 'initial_value': -1.0, 'bounds': [-2, 2]}]

        # multi objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

        # constraints
        self.constraints = [{'name': 'c_1'},
                            {'name': 'c_2'}]

    def evaluate(self, individual):
        x = individual.vector

        f1 = 100 * (x[0] ** 2 + x[1] ** 2)
        f2 = (x[0] - 1) ** 2 + x[1] ** 2

        return [f1, f2]

    def evaluate_inequality_constraints(self, x):
        g1 = 2 * (x[0] - 0.1) * (x[0] - 0.9) / 0.18
        g2 = - 20 * (x[0] - 0.4) * (x[0] - 0.6) / 4.8

        return [g1, g2]


class TestpymooMultiConstraintOptimization(unittest.TestCase):
    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_pymoo_nsgaii(self):
        moo_algorithm = NSGA2(
            pop_size=40,
            n_offsprings=10,
            sampling=FloatRandomSampling(),
            crossover=SimulatedBinaryCrossover(prob_var=0.9, eta=15),
            mutation=PolynomialMutation(eta=20),
            eliminate_duplicates=True
        )

        problem = ProblemConstraint()

        algorithm = Pymoo(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 40
        algorithm.options['algorithm'] = moo_algorithm
        algorithm.run()

        f_1 = []
        f_2 = []
        for individual in problem.last_population():
            f_1.append(individual.costs[0])
            f_2.append(individual.costs[1])

        # print(len(problem.individuals))
        # for individual in problem.individuals:
        #    print(individual)

        self.assertLess(min(f_1), 1.5)
        self.assertGreater(max(f_1), 74)
        self.assertLess(max(f_2), 1.5)
        self.assertGreater(max(f_2), 0.75)

        # import matplotlib.pyplot as plt
        #
        # plt.figure(figsize=(7, 5))
        # # plt.scatter(res.F[:, 0], res.F[:, 1], s=30, facecolors='none', edgecolors='blue')
        # plt.scatter(f_1, f_2, s=30, facecolors='none', edgecolors='blue')
        # plt.title("Objective Space")
        # plt.show()
        #
        # print(min(f_1), max(f_1), min(f_2), max(f_2))


class TestAckley(unittest.TestCase):
    """ Tests that the nsga - ii can find the global optimum. """

    def moo(self, moo_algorithm, population_number=20):
        try:
            problem = Ackley(**{'dimension': 1})

            algorithm = Pymoo(problem)
            algorithm.options['verbose_level'] = 0
            algorithm.options['n_iterations'] = population_number
            algorithm.options['algorithm'] = moo_algorithm
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('f_1')  # Takes last cost function
            #print(optimum)
            self.assertLess(optimum.costs[0], problem.global_optimum + 1.0)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.moo(moo_algorithm, int(1.5 * population_number))

    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_nsga2(self):
        moo_algorithm = NSGA2(
            pop_size=10,
            n_offsprings=10,
            sampling=FloatRandomSampling(),
            crossover=SimulatedBinaryCrossover(prob_var=0.9, eta=15),
            mutation=PolynomialMutation(eta=20),
            eliminate_duplicates=True
        )
        self.moo(moo_algorithm)

    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_isres(self):
        moo_algorithm = ISRES(
            n_offsprings=30,
            rule=1.0 / 7.0,
            gamma=0.85,
            alpha=0.2)
        self.moo(moo_algorithm)

    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_nelder_mead(self):
        moo_algorithm = NelderMead(max_restarts=1)
        self.moo(moo_algorithm)

    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_ga(self):
        moo_algorithm = GA(
            pop_size=20,
            eliminate_duplicates=True)
        self.moo(moo_algorithm)

    @unittest.skipIf(__pymoo__ is False, "requires pymoo")
    def test_de(self):
        moo_algorithm = DE(
            pop_size=20,
            sampling=LHS(),
            variant="DE/rand/1/bin",
            CR=0.3,
            dither="vector",
            jitter=False
        )
        self.moo(moo_algorithm)



import unittest
from ..algorithm_NSGAII import NSGAII
from ..benchmark_functions import Sphere
from ..results import Results

# ToDo: Ask for another goal function
# class Limit_State(Problem):
#
#     def set(self, **kwargs):
#         """Time-dependent 1D QM wave function of a single particle - squared."""
#         self.parameters = [{'name': 'x_1', 'bounds': [0, 100], 'parameter_type': 'float'},
#                            {'name': 'x_2', 'bounds': [0, 100], 'parameter_type': 'float'}]
#         self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]
#
#     def evaluate(self, individual):
#         X = individual.vector
#         x1 = X[0]
#         x2 = X[1]
#         res = x1 ** 3 + x2 ** 3 - 18
#         return [res]


class Test_FORM(unittest.TestCase):
    def test_local_problem(self):
        problem = Sphere(**{'dimension': 1})
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 10
        algorithm.options['max_population_size'] = 10
        algorithm.run()

        result = Results(problem)
        optimal = result.find_optimum('f_1')
        x, reliability_index = result.form_reliability(problem, optimal)
        print(reliability_index)
        self.assertAlmostEqual(x[0], 0.0, places=1)



import numpy as np
from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from artap.operators import CustomGenerator, RandomGenerator, UniformGenerator
import time
from scipy.stats import norm
from .individual import Individual


class Monte_Carlo(GeneticAlgorithm):
    """
    Monte Carlo simulation
        1) make a list of random values of x
        2) pick a small number of random value of x
        3) evaluate the objective function
        4) pick the lowest
    """

    def __init__(self, problem: Problem, name="Monte Carlo"):
        super().__init__(problem, name)
        self.evaluation = 0
        self.z_min, self.z_max = 0, 1
        self.sampling_size = self.options['max_population_size']
        self.problem = problem
        # self.problem.parameters = bm.generate_paramlist(self, dimension=self.dimension, lb=0.0, ub=1.0)
        self.intervals = np.linspace(self.z_min, self.z_max, self.sampling_size)
        self.generator = RandomGenerator(self.problem.parameters)

    def generate(self):
        self.generator.init(self.options['max_population_size'])
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(Individual(vector))
        return individuals

    def run(self):

        individuals = self.generate()

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0
            self.problem.data_store.sync_individual(individual)

        self.evaluate(individuals)

        start = time.time()
        self.problem.logger.info("Monte_Carlo: {}/{}".format(self.options['max_population_number'],
                                                             self.options['max_population_size']))
        for it in range(self.options['max_population_number']):

            individuals = self.generate()
            self.evaluate(individuals)

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - start
        self.problem.logger.info("Monte_Carlo: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()


class Numerical_Integrator(GeneticAlgorithm):
    """
    Adaptive Monte Carlo Integral Approximation.
    Summation is used to estimate the expected value of a function under a distribution, instead of integration.
    """

    def __init__(self, problem: Problem, name="Numerical Integrator using Monte Carlo"):
        super().__init__(problem, name)
        self.evaluation = 0
        self.z_min, self.z_max = 0, 1
        self.a = 0
        self.b = (3*np.pi)/2
        self.problem = problem
        self.num_of_intervals = 10
        self.generator = UniformGenerator(self.problem.parameters)

    def generate(self):
        self.generator.init(self.options['max_population_size'])
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(Individual(vector))
        return individuals

    def run(self):
        individuals = self.generate()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            individual.vector[0] = individual.vector[0] * (self.a + (self.b - self.a))

        self.evaluate(individuals)

        start = time.time()
        self.problem.logger.info("Monte_Carlo: {}/{}".format(self.options['max_population_number'],
                                                             self.options['max_population_size']))
        for it in range(self.options['max_population_number']):

            individuals = []
            individuals = self.generate()

            for individual in individuals:
                individual.vector[0] = individual.vector[0] * (self.a + (self.b - self.a))

            self.evaluate(individuals)

            for individual in individuals:
                # individual.costs[0] = result
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            # self.calculate_integral()
        t = time.time() - start
        self.problem.logger.info("Monte_Carlo: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()


class ImportanceSampling(Monte_Carlo):
    """
    In statistics, importance sampling is a general technique for estimating properties of a particular distribution,
    while only having samples generated from a different distribution than the distribution of interest. The method was
    first introduced by Teun Kloek and Herman K. van Dijk in 1978,[1]

    p(x), given a distribution to be sampled, and a proposal distribution, q(x),
    now define weights w(x) = p(x) / q(x) , such that p(x) = (p(x)/q(x)) * q(x) = w(x)q(x).


    [1] Kloek, T.; van Dijk, H. K. (1978). "Bayesian Estimates of Equation System Parameters: An Application of
        Integration by Monte Carlo". Econometrica. 46 (1): 1–19.

    """

    def __init__(self, problem: Problem, name="Importance Sampling"):
        super().__init__(problem, name)
        self.name = 'Importance Sampling'
        self.min_val, self.max_val = 0, 1
        self.size = self.options['max_population_size']
        self.problem = problem
        self.mean = np.random.uniform(self.min_val, self.max_val, self.size)
        self.std = np.random.uniform(self.max_val - 1, self.max_val, self.size)
        self.cov = np.diag(self.std)

    def generate(self):
        individuals = []
        q = np.random.uniform(self.min_val, self.max_val, self.size)
        q_x = norm(self.mean, self.std).pdf(q)
        p_x = norm(self.min_val, self.max_val).pdf(q)

        weights = p_x / q_x
        samples = weights * q_x
        for i in range(len(samples)):
            individuals.append(Individual([samples[i]], self.individual_features))
        return individuals

    def run(self):
        individuals = self.generate()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            self.problem.data_store.sync_individual(individual)

        start = time.time()
        self.problem.logger.info("ImportanceSampling: {}/{}".format(self.options['max_population_number'],
                                                                    self.options['max_population_size']))
        for it in range(self.options['max_population_number']):

            self.evaluate(individuals)

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - start
        self.problem.logger.info("ImportanceSampling: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()


class Rejection_Sampling(Monte_Carlo):
    """
    Rejection sampling, or the acceptance-rejection method, is one of the basic techniques used to generate
    observations from a distribution. The rejection sampling method generates sampling values from a target
    distribution X with arbitrary probability density function f(x) by using a proposal distribution Y
    with probability density g(x). The idea is that one can generate a sample value from X by instead sampling
    from Y and accepting the sample from Y with probability f(x)/(Cg(x)), repeating the draws from
    Y until a value is accepted. C here is a constant, finite bound on the likelihood ratio f(x)/g(x),
    satisfying  1<=C<infty over the support of X; in other words, C must satisfy f(x)<=Mg(x) for all values of x [1].

    [1] Casella, George; Robert, Christian P.; Wells, Martin T. (2004). Generalized Accept-Reject sampling schemes.
        Institute of Mathematical Statistics. pp. 342–347. doi:10.1214/lnms/1196285403. ISBN 9780940600614.
    """

    def __init__(self, problem: Problem, name="Rejection Sampling"):
        super().__init__(problem, name)
        self.name = 'Rejection Sampling'
        self.min_val, self.max_val = 0, 1
        self.size = self.options['max_population_size']
        self.problem = problem
        self.dimension = 1
        self.mean = np.random.uniform(self.min_val, self.max_val, self.dimension)
        self.std = np.random.uniform(self.max_val - 1, self.max_val, self.dimension)
        self.C = 1.0

    def _p_x(self, x):
        output = norm.pdf(x, self.mean, self.std)
        return output

    def _q_x(self, x):
        output = norm.pdf(x, self.min_val, self.max_val)
        return output

    def generate(self):
        i, samples = 0, []
        individuals = []
        while i < self.size:
            x_i = np.random.normal(self.min_val, self.max_val)
            if self._p_x(x_i) <= self.C * self._q_x(x_i):
                u = np.random.uniform(self.min_val - 1, self.C * self._q_x(x_i))
                if u < self._p_x(x_i):
                    samples.append(x_i)
                    i = i + 1
                else:
                    i = i + 1
        for i in range(len(samples)):
            individuals.append(Individual([samples[i]], self.individual_features))
        return individuals

    def run(self):
        individuals = self.generate()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            self.problem.data_store.sync_individual(individual)
        self.evaluate(individuals)
        start = time.time()
        self.problem.logger.info("Rejection_Sampling: {}/{}".format(self.options['max_population_number'],
                                                                    self.options['max_population_size']))
        for it in range(self.options['max_population_number']):
            individuals = self.generate()
            self.evaluate(individuals)

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - start
        self.problem.logger.info("Rejection_Sampling: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()



import unittest

from artap.problem import Problem
from ..algorithm_monte_carlo import Monte_Carlo, Numerical_Integrator, ImportanceSampling, Rejection_Sampling
from ..results import Results
from ..benchmark_functions import Sphere, Ackley, BenchmarkFunction, Schwefel
import numpy as np


class Integral_Problem(Problem):
    def set(self, **kwargs):
        """Time-dependent 1D QM wave function of a single particle - squared."""
        # self.set_dimension(**kwargs)
        # self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)
        self.parameters = [{'name': 'x1', 'bounds': [0, 1], 'parameter_type': 'integer'}]
        self.u_b = 3 * np.pi / 2
        self.l_b = 0
        self.sampling_size = 100
        self.global_optimum = -1.0

        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]

    def evaluate(self, z):
        x = z.vector
        # output = np.exp(-x[0] ** 2) / (np.pi ** 0.5)
        output = np.sin(x[0])
        return [output]


class TestSphere(unittest.TestCase):
    # unit-test  benchmarck : Sphere, algorithm : Monte_Carlo
    def test_local_problem(self, population_number=50):
        problem = Sphere(**{'dimension': 1})
        algorithm = Monte_Carlo(problem)
        algorithm.options['max_population_number'] = population_number
        algorithm.options['max_population_size'] = 100
        algorithm.run()

        result = Results(problem)
        optimum = result.find_optimum('f_1')
        self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, places=4)


class TestIntegral_MonteCarlo(unittest.TestCase):
    # unit-test  benchmarck : Sphere, algorithm : Integral_Monte_Carlo
    def test_local_problem(self):
        problem = Integral_Problem(**{'dimension': 1})
        algorithm = Numerical_Integrator(problem)
        algorithm.options['max_population_number'] = 50
        algorithm.options['max_population_size'] = 100
        algorithm.options['max_processes'] = 10
        algorithm.run()

        result = Results(problem)
        optimum = result.find_optimum('f_1')
        self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, places=1)


# class TestImpotance_Sampling(unittest.TestCase):
#     # unit-test  benchmarck : Sphere, algorithm : ImportanceSampling
#     def test_local_problem(self):
#         problem = Integral_Problem(**{'dimension': 1})
#         algorithm = ImportanceSampling(problem)
#         algorithm.options['max_population_number'] = 20
#         algorithm.options['max_population_size'] = 10
#         algorithm.run()
#
#         result = Results(problem)
#         optimum = result.find_optimum('f_1')
#         x = "{:0.1f}".format(optimum.costs[0])
#         self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, places=2)
#
#
# class TestRejection_Sampling(unittest.TestCase):
#     # unit-test  benchmarck : Sphere, algorithm : Rejection_Sampling
#     def test_local_problem(self):
#         problem = Sphere(**{'dimension': 1})
#         algorithm = Rejection_Sampling(problem)
#         algorithm.options['max_population_number'] = 20
#         algorithm.options['max_population_size'] = 10
#         algorithm.run()
#         result = Results(problem)
#         optimum = result.find_optimum('f_1')
#         self.assertAlmostEquals(optimum.costs[0], problem.global_optimum, places=2)


if __name__ == '__main__':
    unittest.main()



import unittest

from ..benchmark_functions import Ackley
from ..algorithm_cem import CEM
from ..results import Results


class TestAckley1(unittest.TestCase):
    """ Tests that the CEM can find the global optimum. """

    def test_local_problem(self, population_number=5):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = CEM(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 10
            algorithm.options['max_processes'] = 1
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('f_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


if __name__ == '__main__':
    unittest.main()



import time
from copy import deepcopy
from .algorithm import Algorithm
from .operators import RandomGenerator, SimulatedBinaryCrossover, \
    PmMutator, TournamentSelector, EpsilonDominance, nondominated_truncate, crowding_distance
from .problem import Problem
from .archive import Archive
from .individual import Individual


class GeneticAlgorithm(Algorithm):
    def __init__(self, problem: Problem, name="General Genetic-based Algorithm", evaluator_type=None):

        super().__init__(problem, name, evaluator_type)

        self.parameters_length = len(self.problem.parameters)

        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum evaluations')
        self.options.declare(name='max_population_number', default=100, lower=1,
                             desc='max_population_number')
        self.options.declare(name='max_population_size', default=100, lower=1,
                             desc='Maximal number of individuals in population')

        self.problem = problem
        self.__only_single_objective = False

        # set random generator
        self.generator = RandomGenerator(self.problem.parameters)
        self.generator = None
        self.selector = None
        self.mutator = None
        self.crossover = None

    def generate(self, parents, archive=None):

        offsprings = []
        while len(offsprings) < self.options['max_population_size']:
            parent1 = self.selector.select(parents)

            if archive:
                if len(archive) <= 1:
                    parent2 = self.selector.select(parents)
                else:
                    parent2 = archive.rand_choice()
            else:
                parent2 = self.selector.select(parents)

            # crossover
            vector_1, vector_2 = self.crossover.cross(parent1.vector, parent2.vector)
            child1 = parent1.__class__(vector_1)
            child2 = parent1.__class__(vector_2)
            # mutation
            child1.vector = self.mutator.mutate(child1.vector, child2.vector)
            child2.vector = self.mutator.mutate(child2.vector, child1.vector)

            # always create new individual
            if len(offsprings) == 0:
                offsprings.append(child1)

            if any(child1 == offspring for offspring in offsprings) and (len(offsprings) < self.options[
                'max_population_size']):
                pass
            else:
                offsprings.append(child1)

            if any(child2 == offspring for offspring in offsprings) and (len(offsprings) < self.options[
                'max_population_size']):
                pass
            elif len(offsprings) < self.options['max_population_size']:
                offsprings.append(child2)

        return offsprings

    def run(self):
        pass


class IndividualEpsMOEA(Individual):

    def __init__(self, vector: list):
        super().__init__(vector)
        self.features['dominate'] = []
        self.features['crowding_distance'] = 0
        self.features['domination_counter'] = 0
        self.features['front_number'] = 0
        self.population_id = 0


class EpsMOEA(GeneticAlgorithm):
    """
    EpsMOEA is proposed by Deb et al[1] which is a efficient algorithm with less computer time because it does not
    contain a sort process for non-dominate set. The searching space are divided into several hyper-boxes under
    box dominance concept, when two solutions in same box, only one solution can survive in the hyperbox, EpsMOEA choose
    the solution nearer to the hyperbox corner.

    .. Ref::

    [1] K. Deb, M. Mohan, S. Mishra, Evaluating the epsilon-domination based multiobjective evolutionary algorithm for
        a quick computation of Pareto-optimal solutions, Evolutionary Computation 13(4) (2005) 501 - 525.
    [2] Siwei Jiang, Zhihua Cai: Enhance the Convergence and Diversity for eps-MOPSO by Uniform Design and
                                 Minimum Reduce Hypervolume
    """

    def __init__(self, problem: Problem, name="EpsMOEA Algorithm", evaluator_type=None):
        super().__init__(problem, name, evaluator_type)

        self.options.declare(name='prob_cross', default=0.6, lower=0,
                             desc='prob_cross')
        self.options.declare(name='prob_mutation', default=0.2, lower=0,
                             desc='prob_mutation'),
        self.options.declare(name='epsilons', default=0.01, lower=1e-6,
                             desc='prob_epsilons')

        self.archive = None

    def run(self):
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters)
        self.generator.init(self.options['max_population_size'])
        # set crossover
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])
        self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        # one individual is selected from the archive and one from the population to create a new individual
        self.selector = TournamentSelector(self.problem.parameters)

        # create initial population
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(IndividualEpsMOEA(vector))

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        # an archive to collect the eps-dominated solutions
        self.archive = Archive(dominance=EpsilonDominance(epsilons=self.options['epsilons']))

        # evaluate individuals
        self.evaluator.evaluate(individuals)

        # archiving the eps-dominating solutions
        for individual in individuals:
            self.archive.add(individual)

        # sync to datastore
        for individual in individuals:
            self.problem.data_store.sync_individual(individual)

        t_s = time.time()
        self.problem.logger.info(
            "Eps-MOEA: {}/{}".format(self.options['max_population_number'], self.options['max_population_size']))

        for it in range(self.options['max_population_number']):
            # generate and evaluate the next generation
            offsprings = self.generate(individuals, archive=self.archive)
            self.evaluator.evaluate(offsprings)

            # pop-acceptance procedure, the dominating offsprings will be preserved in the population and  in the archive
            for individual in offsprings:
                # pareto dominated solutions
                self.selector.pop_acceptance(individuals, individual)
                # archived solutions
                self.archive.add(individual)

                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            # make a new population from the previous population
            # individuals = offsprings

        t = time.time() - t_s
        self.problem.logger.info("Eps-MOEA: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()



import numpy as np
from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from .individual import Individual
from .operators import CustomGenerator, nondominated_truncate, RandomGenerator, UniformGenerator
import time


class CMA_ES(GeneticAlgorithm):
    """
    Implementation of CMA_ES, Covariance Matrix Adaptation Evolutionary strategy (CMA_ES).
    The Covariance Matrix Adaptation Evolution Strategy (CMA-ES) [1] is one of the most effective approaches
    for black-box optimization, in which objective functions cannot be specified explicitly in general.
    CMA-ES outperformed over 100 black-box optimization approaches for a variety of benchmark problems [2].

    The CMA-ES algorithm selects solutions from a multivariate gaussian distribution. Following the evaluation of
    all solutions, the solutions are sorted by evaluation values, and the distribution parameters
    (i.e., the mean vector and the covariance matrix) are updated depending on the ranking of evaluation values.

    [1] Nikolaus Hansen and Andreas Ostermeier. Completely derandomized self-adaptation in evolution strategies.
        Evol. Comput., 9(2):159–195, June 2001.
        DOI: http://dx.doi.org/10.1162/106365601750190398.
    [2] Nikolaus Hansen. The CMA Evolution Strategy: A Comparing Review, pages 75–102. Springer Berlin Heidelberg,
        Berlin, Heidelberg, 2006.
        DOI: https://doi.org/10.1007/3-540-32494-1_4.

    """

    def __init__(self, problem: Problem, name="Covariance Matrix Adaptation Evolutionary Strategy"):

        super().__init__(problem, name)
        # Population Size
        self.n_samples = self.options['max_population_size']
        # Number of generation
        self.t = self.options['max_population_number']
        self.individual_features['velocity'] = dict()
        self.individual_features['best_cost'] = dict()
        self.individual_features['best_vector'] = dict()

        self.individual_features['dominate'] = []
        self.individual_features['crowding_distance'] = 0
        self.individual_features['domination_counter'] = 0
        # Add front_number feature
        self.individual_features['front_number'] = 0

        self.dim_theta = len(self.problem.parameters)

        # Elite ratio percentage
        self.top_p = 30
        # Range of values
        self.min_val = 0
        self.max_val = 1
        # Number of Runs
        self.runs = 1
        self.theta_mean = np.random.uniform(self.min_val, self.max_val, self.dim_theta)
        # self.individuals = []
        theta_std = np.random.uniform(self.max_val - 1, self.max_val, self.dim_theta)
        self.theta_cov = np.diag(theta_std)
        self.generator = CustomGenerator(self.problem.parameters)
        # self.fit_gaussian()

    def fit_gaussian(self):
        """
        generates individuals from a multivariate gaussian distribution
        :param
        :return population: list of individuals
        """
        theta = np.random.multivariate_normal(self.theta_mean, self.theta_cov, self.options['max_population_size'])
        individuals = np.clip(theta, self.min_val, self.max_val)
        self.generator.init(individuals)
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(Individual(vector))
        return individuals

    def take_elite(self, candidates):
        """
        Based on the fitness, it will take top individuals
        :param candidates
        :return elite: list of top individuals
        """
        n_top = int((self.n_samples * self.top_p) / 100)
        elite = candidates[:n_top]
        return elite

    def compute_new_mean(self, e_candidates):
        """
        Update distribution parameters. Here, the mean vector will be updated depending on the ranking of
        evaluation values.
        :param e_candidates
        :return new_means vector
        """
        new_means = np.mean(e_candidates, axis=0)
        return new_means

    def compute_new_cov(self, e_candidates):
        """
        Update distribution parameters. Here, the covariance matrix will be updated depending on the ranking of
        evaluation values
        :param e_candidates
        :return new_covariance matrix
        """
        e_candidates = np.array(e_candidates)
        I = np.identity(self.dim_theta)
        cov = np.zeros((self.dim_theta, self.dim_theta))
        for i in range(self.dim_theta):
            for j in range(self.dim_theta):
                cov[i, j] = np.sum(
                    ((e_candidates[:, i] - self.theta_mean[i]) * (e_candidates[:, j] - self.theta_mean[j])), axis=0)

        return 1 / e_candidates.shape[0] * cov + I * 1e-3

    def run(self):
        mean_fitness = []
        best_fitness = []
        worst_fitness = []
        fitness = []
        individuals = self.fit_gaussian()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            self.problem.data_store.sync_individual(individual)

        self.evaluate(individuals)

        start = time.time()
        self.problem.logger.info("CMA_ES: {}/{}".format(self.options['max_population_number'],
                                                        self.options['max_population_size']))
        for it in range(self.options['max_population_number']):
            lists = []
            for individual in individuals:
                # fitness.append(individual.costs)
                lists.append(individual.costs)
            lists = np.array(lists)

            mean_fitness.append(np.mean(lists))
            best_fitness.append(np.min(lists))
            worst_fitness.append(np.max(lists))
            fitness.append(lists)

            elite = self.take_elite(individuals)

            e_candidates = [i.vector for i in elite]

            self.theta_cov = self.compute_new_cov(e_candidates)
            self.theta_mean = self.compute_new_mean(e_candidates)
            individuals = self.fit_gaussian()
            # individuals = nondominated_truncate(new_individuals, self.options['max_population_size'])

            self.evaluate(individuals)

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - start
        self.problem.logger.info("CMA_ES: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()



import unittest

from ..benchmark_functions import AlpineFunction, Ackley
from ..benchmark_pareto import ZDT1
from ..algorithm_NSGAII import NSGAII
from ..results import Results
from ..problem import Problem
from ..quality_indicator import epsilon_add


class TestNSGA2(unittest.TestCase):
    def test_local_problem(self):
        problem = ZDT1()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 10
        algorithm.options['max_population_size'] = 10
        algorithm.options['max_processes'] = 1
        algorithm.run()

        populations = problem.populations()

        # Test if the number of populations is the same as the required number
        self.assertEqual(len(populations), algorithm.options['max_population_number'])

        # Test if it fits the number of individuals in population
        for individuals in populations.values():
            self.assertEqual(len(individuals), algorithm.options['max_population_size'])

        # Test if it fits the overall number of individuals
        self.assertEqual(len(problem.individuals),
                         algorithm.options['max_population_number'] * algorithm.options['max_population_size'])


class TestZDT1(unittest.TestCase):
    # integration test -- tests the total functionality of nsga2
    # around 11secs according to literature DOI: 10.1007/978-3-642-01020-0_39
    def test_local_problem(self):
        problem = ZDT1()
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 100
        algorithm.options['max_population_size'] = 100  # according to the literature
        algorithm.options['max_processes'] = 1
        algorithm.run()

        results = Results(problem)
        vals = results.pareto_values()
        exact = problem.pareto_front(vals[0])
        self.assertLessEqual(epsilon_add(exact, vals), 0.2)


class TestAckley(unittest.TestCase):
    """ Tests that the NSGA - II can find the global optimum. """

    def test_local_problem(self, population_number=50):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['max_processes'] = 1
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('f_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


class TestAlpine(unittest.TestCase):
    """ Tests that the nsga - ii can find the global optimum. """

    def test_local_problem(self, population_number=50):
        try:
            problem = AlpineFunction(**{'dimension': 3})
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['max_processes'] = 1
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('f_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAlpine::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


class ProblemConstraint(Problem):
    def set(self):
        self.name = 'ProblemConstraint'

        self.parameters = [{'name': 'x', 'initial_value': 1.0, 'bounds': [-2, 2]},
                           {'name': 'y', 'initial_value': -1.0, 'bounds': [-2, 2]}]

        # multi objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

        # constraints
        self.constraints = [{'name': 'c_1'},
                            {'name': 'c_2'}]

    def evaluate(self, individual):
        x = individual.vector

        f1 = 100 * (x[0] ** 2 + x[1] ** 2)
        f2 = (x[0] - 1) ** 2 + x[1] ** 2

        return [f1, f2]

    def evaluate_inequality_constraints(self, x):
        g1 = 2 * (x[0] - 0.1) * (x[0] - 0.9) / 0.18
        g2 = - 20 * (x[0] - 0.4) * (x[0] - 0.6) / 4.8

        return [g1, g2]


class MultiConstraintOptimization(unittest.TestCase):
    def test_constraint_nsgaii(self, population_number=10):
        try:
            problem = ProblemConstraint()
            algorithm = NSGAII(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['max_processes'] = 1
            algorithm.run()

            f_1 = []
            f_2 = []
            for individual in problem.last_population():
                f_1.append(individual.costs[0])
                f_2.append(individual.costs[1])

            self.assertLess(min(f_1), 1.5)
            self.assertGreater(max(f_1), 74)
            self.assertLess(max(f_2), 1.5)
            self.assertGreater(max(f_2), 0.75)

        # ToDo: Is it good in test?
        except AssertionError:
            # stochastic
            print("MultiConstraintOptimization::test_local_problem", population_number)
            self.test_constraint_nsgaii(int(1.5 * population_number))


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import numpy as np
from ..executor import LocalCSTExecutor
from ..problem import Problem
from ..algorithm import DummyAlgorithm
from ..individual import Individual
from .tests_root import tests_root_path

# Test if the Cst software is installed on the system
# Note: cst must be in system path
__cst__ = True
result = os.system('"CST DESIGN ENVIRONMENT_AMD64.exe" /version')
if result != 0:
    __cst__ = False


class CstProblem(Problem):
    #
    def set(self):
        self.name = "ComsolProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10},
                           {'name': 'b', 'initial_value': 10}]
        self.costs = [{'name': 'F1', 'criteria': 'minimize'}]
        self.output_files = ["patch_circular_polarization\\Export\\Farfield\\farfield_source_broadband_[1].ffs"]
        problem_file = os.path.join(tests_root_path, 'data\\cst\\patch_circular_polarization.cst')
        self.executor = LocalCSTExecutor(self, model_file=problem_file)

    #
    def evaluate(self, individual):
        individual.dep_param = 0
        return self.executor.eval(individual)

    def parse_results(self, individual, output_files=None):
        # output_file = output_files[0]
        # path = output_file
        #
        # content = ""
        # if os.path.exists(path):
        #     with open(path) as file:
        #         content = file.read()
        # else:
        #     self.logger.error(
        #         "Output file '{}' doesn't exists.".format(path))
        # far_field = np.zeros([361, 181])
        # with open(output_file) as file:
        #     lines = file.readlines()
        #
        # for line in lines[31:]:
        #     data = line.split()
        #     numbers = []
        #     for item in data:
        #         numbers.append(float(item))
        #
        #     phi = int(numbers[0])
        #     theta = int(numbers[1])
        #     x = np.abs(numbers[2] + 1j * numbers[3])
        #     y = np.abs(numbers[4] + 1j * numbers[5])
        #     far_field[phi, theta] = np.sqrt(x ** 2 + y ** 2)

        return 10


class TestLocalComsol(unittest.TestCase):
    @unittest.skipIf(__cst__ is False, "require CST")
    def test_cst_exec(self):
        """ Tests one calculation of goal function."""
        problem = CstProblem()

        individuals = [Individual([10, 10])]
        algorithm = DummyAlgorithm(problem)
        algorithm.evaluator.evaluate(individuals)
        try:
            algorithm.evaluate(individuals)
            self.assertAlmostEqual(112.94090668383139, individuals[0].costs[0])
        except Exception as e:
            print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            print(e)



from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from .operators import GradientEvaluator, RandomGenerator
from .individual import Individual
import time
import math


class GradientDescent(GeneticAlgorithm):
    """ Gradient descent algorithms """

    def __init__(self, problem: Problem, generator=None, name="GradientDescent"):
        super().__init__(problem, name)
        self.problem = problem

        self.options.declare(name='n_iterations', default=10,
                             desc='Maximum number of iterations')
        self.options.declare(name='step', default=0.1,
                             desc='Step')
        self.options.declare(name='minimal_step', default=1e-6,
                             desc='Minimal step')
        self.options.declare(name='algorithm', default="fixed",
                             desc='Type of algorithm: fixed, adaptive, adagrad, rmsprop, adam')
        self.options.declare(name='c', default=1e-4,
                             desc='Armijo coefficient (Adaptive)')
        self.options.declare(name='beta1', default=1e-4,
                             desc='Coefficient beta 1 (Adam)')
        self.options.declare(name='beta2', default=1e-4,
                             desc='Coefficient beta 2 (Adam)')
        self.options.declare(name='decay_rate', default=1e-4,
                             desc='Decay Rate (RMSprop)')

        # compute gradients
        self.evaluator = GradientEvaluator(self)
        self.individual_features['gradient'] = dict()

        # set default generator
        if generator is None:
            self.generator = RandomGenerator(self.problem.parameters)
            self.generator.init(1)
        else:
            self.generator = generator

    def step_fixed(self, individual):
        x = []
        for i in range(len(individual.vector)):
            x.append(individual.vector[i] - self.options["step"] * individual.features['gradient'][i])

        return x

    def step_adaptive(self, individual):
        # Armijo line search
        h_t = self.options["step"]
        c = self.options["c"]

        grad_sqr = 0
        for i in range(len(individual.vector)):
            grad_sqr = grad_sqr + individual.features['gradient'][i] ** 2

        x_static = []
        for i in range(len(individual.vector)):
            x_static.append(individual.vector[i])
        individual_static = Individual(x_static)
        self.evaluate([individual_static])

        while True:
            x_armijo = []
            for i in range(len(individual.vector)):
                x_armijo.append(individual.vector[i] - h_t * individual.features['gradient'][i])
            individual_armijo = Individual(x_armijo)
            self.evaluate([individual_armijo])

            satisfied = True
            for i in range(len(individual.costs)):
                if individual_armijo.costs[i] > individual_static.costs[i] - c * h_t * grad_sqr:
                    satisfied = False
            if satisfied:
                return x_armijo

            # step too small
            if h_t < self.options["minimal_step"]:
                return x_armijo

            h_t = h_t / 2.0

    def step_adagrad(self, individual):
        x = []
        for i in range(len(individual.vector)):
            self.cache_grad_sqr[i] += individual.features['gradient'][i] ** 2

            x.append(individual.vector[i] - self.options["step"] * individual.features['gradient'][i] /
                     (math.sqrt(self.cache_grad_sqr[i] + 1e-6)))

        return x

    def step_rmsprop(self, individual):
        x = []
        for i in range(len(individual.vector)):
            self.cache_grad_sqr[i] += self.options["decay_rate"] * self.cache_grad_sqr[i] \
                                      + (1.0 - self.options["decay_rate"]) * individual.features['gradient'][i] ** 2

            x.append(individual.vector[i] - self.options["step"] * individual.features['gradient'][i] /
                     (math.sqrt(self.cache_grad_sqr[i] + 1e-6)))

        return x

    def step_adam(self, individual):
        x = []
        for i in range(len(individual.vector)):
            self.adam_t += 1

            self.cache_grad[i] += self.options["beta1"] * self.cache_grad[i] \
                                  + (1.0 - self.options["beta1"]) * individual.features['gradient'][i]
            self.cache_grad_sqr[i] += self.options["beta2"] * self.cache_grad_sqr[i] \
                                      + (1.0 - self.options["beta2"]) * individual.features['gradient'][i] ** 2

            # calculates the bias-corrected estimates
            m_corrected = self.cache_grad[i] / (1.0 - self.options["beta1"] ** self.adam_t)
            v_corrected = self.cache_grad_sqr[i] / (1.0 - self.options["beta2"] ** self.adam_t)

            x.append(individual.vector[i] - self.options["step"] * m_corrected / (math.sqrt(v_corrected + 1e-8)))

        return x

    def generate(self):
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(Individual(vector))
        return individuals

    def run(self):
        # optimization
        t_s = time.time()
        self.problem.logger.info("GradientDescent")

        # cache
        self.adam_t = 0
        self.cache_grad = [0] * len(self.problem.parameters)
        self.cache_grad_sqr = [0] * len(self.problem.parameters)

        # generate initial point
        individuals = self.generate()

        # append to problem
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        self.evaluate(individuals)

        for j in range(self.options["n_iterations"]):
            new_individuals = []
            for individual in individuals:
                # get new position
                if self.options["algorithm"] == "fixed":
                    x = self.step_fixed(individual)
                elif self.options["algorithm"] == "adaptive":
                    x = self.step_adaptive(individual)
                elif self.options["algorithm"] == "adagrad":
                    x = self.step_adagrad(individual)
                elif self.options["algorithm"] == "rmsprop":
                    x = self.step_rmsprop(individual)
                elif self.options["algorithm"] == "adam":
                    x = self.step_adam(individual)
                else:
                    raise "Algorithm '{}' is defined.".format(self.options["algorithm"])

                # create a new individual
                individual = Individual(x)
                individual.population_id = j + 1
                new_individuals.append(individual)

                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            self.evaluate(new_individuals)
            individuals = new_individuals

        t = time.time() - t_s
        self.problem.logger.info("GradientDescent: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()



import numpy as np
from .individual import Individual
from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from artap.operators import CustomGenerator, RandomGenerator, UniformGenerator
import time


class CEM(GeneticAlgorithm):
    """Cross-Entropy Method
    https://link.springer.com/content/pdf/10.1007/s10479-005-5724-z.pdf
    """

    def __init__(self, problem: Problem, name="Cross-Entropy Method"):
        super().__init__(problem, name)
        self.problem = problem
        # Initialize mean and standard deviation
        # self.theta_mean = np.zeros((dim_theta))
        self.n_samples = self.options['max_population_size']
        # Number of generation
        self.t = self.options['max_population_number']
        self.individual_features['velocity'] = dict()
        self.individual_features['best_cost'] = dict()
        self.individual_features['best_vector'] = dict()

        self.individual_features['dominate'] = []
        self.individual_features['crowding_distance'] = 0
        self.individual_features['domination_counter'] = 0
        # Add front_number feature
        self.individual_features['front_number'] = 0

        self.dim_theta = len(self.problem.parameters)

        # Elite ratio percentage
        self.top_p = 30

        # Range of values
        self.min_val = 0
        self.max_val = 1
        self.theta_mean = np.random.uniform(self.min_val, self.max_val, (self.n_samples, self.dim_theta))
        self.theta_std = np.random.uniform(self.min_val, self.max_val, (self.n_samples, self.dim_theta))
        self.generator = CustomGenerator(self.problem.parameters)

    def fit_gaussian(self):
        generation = np.random.normal(self.theta_mean, self.theta_std)
        individuals = np.clip(generation, self.min_val, self.max_val)
        self.generator.init(individuals)
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            individuals.append(Individual(vector))
        return individuals

    def take_elite(self, candidates):
        n_top = int((self.n_samples * self.top_p) / 100)
        elite = candidates[:n_top]
        return elite

    def compute_new_mean(self, e_candidates):
        new_means = np.mean(e_candidates, axis=0)
        new_means = np.tile(new_means, (self.n_samples, 1))
        return new_means

    def compute_new_std(self, e_candidates):
        eps = 1e-3
        new_std = np.std(e_candidates, ddof=1, axis=0) + eps
        return new_std

    def run(self):
        mean_fitness = []
        best_fitness = []
        worst_fitness = []
        fitness = []
        individuals = self.fit_gaussian()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            self.problem.data_store.sync_individual(individual)

        self.evaluate(individuals)

        start = time.time()
        self.problem.logger.info("CEM: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))
        for it in range(self.options['max_population_number']):

            lists = []
            for individual in individuals:
                lists.append(individual.costs)
            lists = np.array(lists)

            mean_fitness.append(np.mean(lists))
            best_fitness.append(np.min(lists))
            worst_fitness.append(np.max(lists))
            fitness.append(lists)

            elite = self.take_elite(individuals)

            e_candidates = [i.vector for i in elite]

            self.theta_mean = self.compute_new_mean(e_candidates)
            self.theta_std = self.compute_new_std(e_candidates)
            individuals = self.fit_gaussian()

            self.evaluate(individuals)
            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - start
        self.problem.logger.info("CEM: elapsed time: {} s".format(t))
        # sync changed individual informations
        self.problem.data_store.sync_all()



import copy
import time
from abc import abstractmethod

from .individual import Individual
from .operators import RandomGenerator, SimulatedBinaryCrossover, \
    PmMutator, TournamentSelector, EpsilonDominance, nondominated_truncate, crowding_distance
from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm


class IndividualNSGAII(Individual):

    def __init__(self, vector: list):
        super().__init__(vector)
        self.features['dominate'] = []
        self.features['crowding_distance'] = 0
        self.features['domination_counter'] = 0
        self.features['front_number'] = 0
        self.population_id = 0

    def copy(self):
        new_individual = self.__class__(self.vector)
        new_individual.costs = self.costs
        new_individual.costs_signed = self.costs_signed
        return new_individual


class NSGAII(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="NSGA_II Evolutionary Algorithm", evaluator_type=None):
        """
         NSGA-II implementation as described in

         [1] K. Deb, A. Pratap, S. Agarwal and T. Meyarivan, "A fast and elitist
             multiobjective genetic algorithm: NSGA-II," in IEEE Transactions on Evolutionary Computation,
             vol. 6, no. 2, pp. 182-197, Apr 2002. doi: 10.1109/4235.996017
        """
        super().__init__(problem, name, evaluator_type)

        self.options.declare(name='prob_cross', default=1.0, lower=0,
                             desc='prob_cross')

        self.options.declare(name='prob_mutation', default=1.0 / (len(problem.parameters)), lower=0,
                             desc='prob_mutation')

    def run(self):
        if self.generator is None:
            self.generator = RandomGenerator(self.problem.parameters)
            self.generator.init(self.options['max_population_size'])
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])
        self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        self.selector = TournamentSelector(self.problem.parameters)

        # create initial population
        vectors = self.generator.generate()
        individuals = []
        for vector in vectors:
            # append to problem
            individuals.append(IndividualNSGAII(vector))
            # add to population

        # evaluate
        self.evaluate(individuals)

        # non-dominated sort of individuals
        self.selector.fast_nondominated_sorting(individuals)

        # sync to datastore
        for individual in individuals:
            self.problem.individuals.append(individual)
            individual.population_id = 1
            self.problem.data_store.sync_individual(individual)

        t_s = time.time()
        self.problem.logger.info(
            "NSGA_II: {}/{}".format(self.options['max_population_number'],
                                    self.options['max_population_number'] * self.options['max_population_size']))

        # optimization
        for it in range(self.options['max_population_number']-1):

            # generate new offsprings
            offsprings = self.generate(individuals)

            # Todo: this lead to too many calculations of the goal function
            # evaluate the offsprings
            self.evaluate(offsprings)

            for individual in individuals:
                offsprings.append(individual.copy())

            # make the pareto dominance calculation and calculating the crowding distance
            self.selector.fast_nondominated_sorting(offsprings)

            # truncate
            # ToDO: Deside if we want to save removed individuals
            # individuals, removed = nondominated_truncate(offsprings, self.options['max_population_size'])
            individuals = nondominated_truncate(offsprings, self.options['max_population_size'])
            for individual in individuals:
                # add to population
                individual.population_id = it + 2
                # append to problem
                self.problem.individuals.append(individual)
                self.problem.data_store.sync_individual(individual)


        t = time.time() - t_s
        self.problem.logger.info("NSGA_II: elapsed time: {} s".format(t))

        # sync changed individual informations
        self.problem.data_store.sync_all()



import csv
import numpy as np
from .quality_indicator import gd, epsilon_add
from .operators import derivative, std_linear
from .individual import Individual


class Results:
    """ Class Results offers tools for interpreting calculated data. """

    def __init__(self, problem):
        self.problem = problem

    def parameter_names(self):
        """ Returns a list of parameter names for the given problem. """
        parameter_names = []
        for parameter in self.problem.parameters:
            parameter_names.append(parameter["name"])
        return parameter_names

    def parameter_number(self):
        """ Returns the number of parameters."""
        return len(self.problem.parameters)

    def goal_number(self):
        """ Returns the number of goal functions. """
        return len(self.problem.costs)

    def goal_names(self):
        """ Returns a list of goal names for the given problem. """
        cost_names = []
        for cost in self.problem.costs:
            cost_names.append(cost["name"])
        return cost_names

    def parameter_index(self, name):
        try:
            return next((index for (index, d) in enumerate(self.problem.parameters) if d["name"] == name), None)
        except ValueError as e:
            raise ValueError('There is not a parameter with given name: {}'.format(name))

    # ToDo: Replace exception raise by warning?
    def goal_index(self, name):
        index = next((index for (index, d) in enumerate(self.problem.costs) if d["name"] == name), None)
        if index is not None:
            return index
        else:
            raise ValueError('There is not a cost function with given name: {}'.format(name))

    def parameters(self):
        out = []
        for individuals in self.problem.populations().values():
            for individual in individuals:
                out.append(individual.vector)
        return out

    def costs(self):
        """ Returns the list containing cost values for all individuals.
            Example of usage:
                costs = results.costs()
                plt.plot(costs, '.')
        """
        out = []
        n = len(self.problem.individuals[0].costs)
        for i in range(n):
            out.append([])

        for individual in self.problem.individuals:
            for i in range(n):
                out[i].append(individual.costs[i])
        return out

    def table(self, transpose=True):
        """ Returns the table where rows consist of values of parameters and values of goal functions. """
        out = []
        for individuals in self.problem.populations().values():
            for individual in individuals:
                out.append(individual.vector + individual.costs)

        if transpose:
            return list(zip(*out))
        else:
            return out

    # TODO : This method returns all the populations instead of the population in the given index
    def population(self, population_id=-1):
        # Returns population with given index
        #   if population_id >= len(self.problem.populations):
        #    raise ValueError('Index of population exceeds the number of populations.')
        if population_id == -1:
            # TODO : last_population returns all the populations instead of last index population
            individuals = self.problem.last_population()
        else:
            # TODO : population() must be modified to return the given index population
            individuals = self.problem.population(population_id)
        return individuals

    # I assume this method must return the goal on the given index not to return all the population or the name must
    # change
    def goal_on_index(self, name=None, population_id=-1):
        """
        Returns a list of lists. The first list contains indexes of individuals in population,
        the other lists contains values of the goal function(s).

        """
        individuals = self.population(population_id)
        n = range(len(individuals))

        table = [list(n)]
        if name is None:
            for j in range(len(self.problem.costs)):
                table.append([])
                for i in n:
                    table[j + 1].append(individuals[i].costs[j])
        else:
            table.append([])
            index = self.goal_index(name)

            for i in n:
                table[1].append(individuals[i].costs[index])

        return table

    # I assume this method must return the goal on the given index not to return all the population or the name must
    # change
    def parameter_on_index(self, name=None, population_id=-1):
        """
        Returns a list of lists. The first list contains indexes of individuals in population,
            the other lists contains values of the parameters(s).
        :param name: string representing the name of parameter if it is not given, all parameters are included
        :param population_id: index of the selected parameter
        :return:
        """

        individuals = self.population(population_id)
        n = range(len(individuals))
        table = [list(n)]
        if name is None:
            for j in range(len(self.problem.parameters)):
                table.append([])
                for i in n:
                    table[j + 1].append(individuals[i].vector[j])
        else:
            table.append([])
            index = self.parameter_index(name)
            for i in n:
                table[1].append(individuals[i].vector[index])

        return table

    def goal_on_parameter(self, parameter_name, goal_name, population_id=-1, sorted=False):
        """
        The method returns the dependance of selected goal function on particular parameter
        :param parameter_name: string specifying particular parameter
        :param goal_name: string specifying name of required goal function
        :param population_id: index of required population, if it is not given the last population is used
        :return: list of two lists, the first contains sorted parameter values, the second values of selected goal
        function
        """

        if population_id == -1:
            individuals = self.problem.last_population()
        else:
            individuals = self.population(population_id)

        parameter_index = self.parameter_index(parameter_name)
        goal_index = self.goal_index(goal_name)
        parameter_values = []
        goal_values = []

        for individual in individuals:
            parameter_values.append(individual.vector[parameter_index])
            goal_values.append(individual.costs[goal_index])

        if sorted is True:
            goal_values = self.sort_list(parameter_values, goal_values)
            parameter_values.sort()

        return [parameter_values, goal_values]

    def parameter_on_goal(self, goal_name, parameter_name, population_id=-1, sorted=False):
        parameter_values, goal_values = self.goal_on_parameter(parameter_name, goal_name, population_id, sorted=False)
        if sorted:
            parameter_values = self.sort_list(goal_values, parameter_values)
            goal_values.sort()

        return [goal_values, parameter_values]

    def parameter_on_parameter(self, parameter_1, parameter_2, population_id=-1, sorted=False):
        individuals = self.population(population_id)
        index_1 = self.parameter_index(parameter_1)
        index_2 = self.parameter_index(parameter_2)
        values_1 = []
        values_2 = []

        for individual in individuals:
            values_1.append(individual.vector[index_1])
            values_2.append(individual.vector[index_2])

        if sorted:
            values_2 = self.sort_list(values_1, values_2)
            values_1.sort()

        return [values_1, values_2]

    def export_to_csv(self, filename):
        """
        Writes out populations into a csv file, the file contains the costs and the parameters
        :return:
        """
        with open(filename, 'w', newline='') as f:
            writer = csv.writer(f)
            # names
            out = ["population_id"]
            for parameter in self.problem.parameters:
                out.append(parameter["name"])
            for cost in self.problem.costs:
                out.append(cost["name"])

            writer.writerows([out])

            # values
            population_id = 0
            for individuals in self.problem.populations().values():
                for individual in individuals:
                    out = [population_id]
                    for i in individual.costs:
                        out.append(i)
                    for j in individual.vector:
                        out.append(j)

                    writer.writerows([out])
                    # population_id += 1
                    # increasing population_id does not work properly here, it must be written as below
                    population_id += 1

    def pareto_individuals(self, population_id=None):
        """

        :return: pareto front
        """
        if population_id is not None:
            individuals = self.population(population_id)
        else:
            individuals = self.population(-1)

        pareto_individuals = []
        for individual in individuals:
            if individual.features['front_number'] == 1:
                pareto_individuals.append(individual)

        return pareto_individuals

    def pareto_front(self, population_id=None):
        """

        :return: pareto front
        """
        if population_id is not None:
            individuals = self.population(population_id)
        else:
            individuals = self.population(-1)

        n = self.goal_number()
        pareto_front = []
        for i in range(n):
            pareto_front.append([])

        for individual in individuals:
            if individual.features['front_number'] == 1:
                for i in range(n):
                    pareto_front[i].append(individual.costs[i])

        return pareto_front

    @staticmethod
    def sort_list(list_1, list_2):
        zipped_pairs = zip(list_1, list_2)
        sorted_list = [x for _, x in sorted(zipped_pairs)]
        return sorted_list

    # TODO: Generalize
    # TODO: add test
    def find_optimum(self, name=None):
        """
        Search the optimal value for the given (by name parameter) single objective .

        :param name: string representing name of the goal function
        :return:
        """

        # get the index of the required parameter
        index = 0  # default - one parameter
        min_l = []
        if name:
            index = self.goal_index(name)

        criteria = None
        if 'criteria' in self.problem.costs[index]:
            criteria = self.problem.costs[index]['criteria']

        if criteria == 'minimize' or criteria is None:
            if len(self.problem.individuals) > 0:
                min_l = [min(self.problem.individuals, key=lambda x: x.costs[index])]
        else:
            if len(self.problem.individuals) > 0:
                min_l = [max(self.problem.individuals, key=lambda x: x.costs[index])]

        # for population in self.problem.populations:
        opt = min(min_l, key=lambda x: x.costs[index])
        return opt

    # TODO: same function - David, could you check it and write test?
    def pareto_values(self, archive=None):
        """
        :return: a list of lists which contains the optimal values of the cost function:
                l_sol[[c11, c12, ... c1n], ... [cm1, cm2, ... cmn]]
        """

        individuals = self.problem.last_population()
        l_sol = []

        # the pareto values collected in the archive, if the algorithm uses this strategy
        if archive:
            for individual in archive:
                l_sol.append(individual.costs)
        else:
            if len(individuals) > 1:
                for individual in individuals:
                    l_sol.append(individual.costs)

        return l_sol

    # TODO: add test
    def performance_measure(self, reference: list, type='epsilon'):
        """
        This function compares the result with a given, reference pareto-set. By default it offers to make
        an epsilon measure, but other metrics can be selected.

        :param reference: list of tuples, which are similar to the data of the calculated pareto-front
        :param type:
        :return:
        """

        result = 0
        computed = self.pareto_values()  # [[], [], .. ]

        if type == 'epsilon':
            result = epsilon_add(reference, computed)
        if type == 'gd':
            result = gd(reference, computed)

        return result

    def get_population_ids(self):
        ids = set()
        for individual in self.problem.individuals:
            ids.add(individual.population_id)
        return ids

    def integration_measure(self):
        Integration = 0
        opt = self.find_optimum('f_1')
        for individual in self.problem.individuals:
            Integration += individual.costs[0]
        ub = self.problem.u_b
        lb = self.problem.l_b
        sampling_size = self.problem.sampling_size
        result = ((ub - lb) / len(self.problem.individuals)) * Integration

        return result

    def form_reliability(self, problem, individual):
        """
            Performs the FORM-HLRF algorithm
                The goal is to find the reliability index for a nonlinear function g(Xi).

                HL_RF algorithm steps:
                1) Define the limit state function (𝑔) and the convergence criterion (tol = 1e-3), statistical characteristics
                 of basic random variables (mean, standard deviation, and distribution function for each random variable).
                2) Transfer the random variables from 𝑋-space to 𝑈-space according to
                    .. math::
                        U = (X - \mu_X^{e}) / \sigma_X^{e}
                3) Compute the gradient vector and limit state function at point U𝑘.
                4) Compute the cost function using
                    .. math::
                        U^{k+1} = [(G(U^{k}) - \grad G(U^{k})U^{k}) / \norm{\grad G(U^{k})}] \alpha^{k}

                    where \alpha^{k} is sensivity factor that can be calculated by:
                    ..math::
                        \alpha^{k} = -(\grad G(U^{k})) / \norm{\grad G(U^{k})}

                5) Transfer the basic random variables from 𝑈-space to 𝑋-space.
                6) Control the convergence conditions as below; if true then the method is converged, stop and go to step (7);
                else, 𝑘 = 𝑘+1 and go to step (2).
                7) Estimate the failure probability by
                    ..math::
                        P_f ~~ \phi(-\beta)
                    where \phi is the CDF of the standard normal distribution and \beta = \norm{U^{*}} is the reliability index.

                ..Reference::

                [1] A Novel Algorithm for Structural Reliability Analysis Based on Finite Step Length and Armijo Line Search
                    https://www.mdpi.com/2076-3417/9/12/2546
                [2] A Comparative Study of First-Order Reliability Method-Based Steepest Descent Search Directions for
                 Reliability Analysis of Steel Structures, https://www.hindawi.com/journals/ace/2017/8643801/
                [3] Reliability analysis and optimal design under uncertainty - Focus on adaptive surrogate-based approaches
                    https://tel.archives-ouvertes.fr/tel-01737299/document

            Args:
                problem
                individual which is optimal
            Return:
                vector: converged design point
                reliability_index
        """
        convergence = False
        tol = 1e-3
        mean = np.random.uniform(0, 10, len(problem.parameters))
        std = np.random.uniform(0, 10, len(problem.parameters))

        mu_g = problem.evaluate(individual)
        grad_g_x = np.array(derivative(problem, mean))
        sig_g = std_linear(grad_g_x, std)
        # reliability_index
        reliability_index = mu_g / sig_g
        alpha = - grad_g_x * std / sig_g

        # Initialize design points
        vector = mean + reliability_index * std * alpha

        while not convergence:
            transform_individuals = (vector - mean) / std

            # compute gradient of g with respect to z
            grad_g_x = derivative(problem, vector)
            grad_g_z = grad_g_x * std

            beta_previous = reliability_index
            reliability_index = (problem.evaluate(Individual(vector)) - grad_g_z @ transform_individuals) \
                                / np.linalg.norm(grad_g_z)
            alpha = - grad_g_z / np.linalg.norm(grad_g_z)

            # update design points in standard space
            transform_individuals = alpha * reliability_index
            # transform to physical space
            physical_transform = mean + transform_individuals * std

            condition1 = np.linalg.norm(physical_transform - vector) \
                         / np.linalg.norm(physical_transform) < tol
            condition2 = abs(np.round(reliability_index, 3) - np.round(beta_previous, 3)) < tol
            if condition2 or condition1:
                convergence = True
            if not convergence:
                vector = physical_transform

        return vector, reliability_index



import time

from .algorithm import Algorithm
from .algorithm_genetic import GeneralEvolutionaryAlgorithm
from .individual import Individual
from .operators import CustomGenerator
from .problem import Problem
import numpy as np
import scipy.special as sp
from scipy import stats

"""
Distributions
"""


class Distribution(object):
    def __init__(self, name=None, mean=None, std=None, startpoint=None):
        self.name = name
        self.mean = mean
        self.std = std
        if startpoint is None:
            self.startpoint = self.mean
        else:
            self.startpoint = startpoint

    def pdf(self, x):
        pass

    def cdf(self, x):
        pass


class Normal(Distribution):
    """Normal distribution
    """

    def __init__(self, name, mean, std):
        self.name = name
        self.mean = mean
        self.std = std
        Distribution.__init__(self, name, mean, std)

    def pdf(self, x):
        """
        probability density function
        """
        z = (x - self.mean) / self.std
        p = np.exp(-0.5 * z ** 2) / np.sqrt(2 * np.pi) / self.std
        return p

    def cdf(self, x):
        """
        cumulative distribution function
        """
        z = (x - self.mean) / self.std
        p = 0.5 + sp.erf(z / np.sqrt(2)) / 2
        return p

    def inv_cdf(self, x):
        """
        inverse cumulative distribution function
        """
        z = sp.erfinv(2 * x - 1) * np.sqrt(2)
        p = self.std * z + self.mean
        return p


class Uniform(Distribution):
    """Uniform distribution
    """

    def __init__(self, name, mean, std):

        self.name = name
        self.mean = mean
        self.std = std
        self.a = self.mean - 3 ** 0.5 * self.std
        self.b = self.mean + 3 ** 0.5 * self.std

        Distribution.__init__(self, name, mean, std)

    def pdf(self, x):
        """probability density function
        """
        if len(x) == 1:
            p = 1 * (self.b - self.a) ** (-1)
        else:
            p = []
            for i in range(len(x)):
                p.append(1 * (self.b - self.a) ** (-1))

        return p

    def cdf(self, x):
        """cumulative distribution function
        """
        P = (x - self.a) * (self.b - self.a) ** (-1)
        return P


class StochasticModel(object):
    """Creates an object with random variable attributes
    Args:
        list with: [distribution, mean, std]
    """

    def __init__(self, *args):
        self.dist_func = []
        mean = []
        std = []
        self.dist_name = []

        for dist, mu, sig in args:

            if dist == 'norm':
                self.dist_func.append(getattr(stats, dist)(loc=mean, scale=sig))

            elif dist == 'uniform':
                self.dist_func.append(getattr(stats, dist)(loc=mean, scale=sig))

            self.dist_name.append(dist)
            mean.append(mu)
            std.append(sig)

        self.mean = np.array(mean)
        self.std = np.array(std)
        """ Correlation Matrix"""
        self.rho = np.identity(len(args))


"""
Reliability Analysis Methods
"""


# ToDo: Convert this class to the optimization class.
class FORM(GeneralEvolutionaryAlgorithm):
    """Performs the FORM-HLRF algorithm
        The goal is to find the reliability index for a nonlinear limit state function g(Xi).

        HL_RF algorithm steps:
        1) Define the limit state function (𝑔) and the convergence criterion (tol = 1e-3), statistical characteristics
         of basic random variables (mean, standard deviation, and distribution function for each random variable).
        2) Transfer the random variables from 𝑋-space to 𝑈-space according to
            .. math::
                U = (X - \mu_X^{e}) / \sigma_X^{e}
        3) Compute the gradient vector and limit state function at point U𝑘.
        4) Compute the cost function using
            .. math::
                U^{k+1} = [(G(U^{k}) - \grad G(U^{k})U^{k}) / \norm{\grad G(U^{k})}] \alpha^{k}

            where \alpha^{k} is sensivity factor that can be calculated by:
            ..math::
                \alpha^{k} = -(\grad G(U^{k})) / \norm{\grad G(U^{k})}

        5) Transfer the basic random variables from 𝑈-space to 𝑋-space.
        6) Control the convergence conditions as below; if true then the method is converged, stop and go to step (7);
        else, 𝑘 = 𝑘+1 and go to step (2).
        7) Estimate the failure probability by
            ..math::
                P_f ~~ \phi(-\beta)
            where \phi is the CDF of the standard normal distribution and \beta = \norm{U^{*}} is the reliability index.

        ..Reference::

        [1] A Novel Algorithm for Structural Reliability Analysis Based on Finite Step Length and Armijo Line Search
            https://www.mdpi.com/2076-3417/9/12/2546
        [2] A Comparative Study of First-Order Reliability Method-Based Steepest Descent Search Directions for
         Reliability Analysis of Steel Structures, https://www.hindawi.com/journals/ace/2017/8643801/
        [3] Reliability analysis and optimal design under uncertainty - Focus on adaptive surrogate-based approaches
            https://tel.archives-ouvertes.fr/tel-01737299/document
    """

    def __init__(self, problem: Problem, name="First Order Reliability Method"):
        super().__init__(problem, name)
        self.min_val, self.max_val = 1, 10
        # self.size = self.options['max_population_size']
        self.dimension = len(self.problem.parameters)
        self.problem = problem
        self.mean = np.random.uniform(self.min_val, self.max_val, self.dimension)
        self.std = np.random.uniform(self.max_val - 1, self.max_val, self.dimension)
        self.cov = np.diag(self.std)
        self.tol = 1e-3
        self.generator = CustomGenerator(self.problem.parameters, self.individual_features)

    def generate(self):
        # X = StochasticModel(['norm', int(self.mean), int(self.std)],
        #                     ['norm', int(self.mean), int(self.std)])
        new_individuals = np.random.multivariate_normal(self.mean, self.cov, self.options['max_population_size'])
        self.generator.init(new_individuals)
        individuals = self.generator.generate()
        return individuals

    def make_individual(self):
        self.mean = Individual(self.mean, self.individual_features)
        self.std = Individual(self.std, self.individual_features)

    def derivative(self, problem, points):
        """Compute derivative of func at points using finite differences

           ddx = \\frac{func(points + eps) - func(points - eps)}{2 * eps}

        Args:
            problem (function): function with N parameters
            points (array): array with N-dimension
        Returns:
            derivative: list with derivative
            for instance:
                derivative[0] = d/d X1 func evaluated at points[0]
                derivative[1] = d/d X2 func evaluated at points[1]
        """
        d = []
        eps = 1e-8
        for i, p in enumerate(points.vector):
            step_up, step_down = points, points  # copy list

            # approximate derivative by tangent line with eps distance at
            step_up.vector[i] = p + eps
            step_down.vector[i] = p - eps
            d.append((problem.evaluate(step_up)[0] - problem.evaluate(step_down)[0]) / (2 * eps))

        return np.array(d)

    def std_linear(self, grad_g_x, sig):
        """Compute the standard deviation of g(Xi) using linear approximation
            . math::
                sig_g = (sum_i^n (grad_g_xi * sig_Xi)^2)^(1/2)
        Args:
            grad_g_x (array): partial derivative of g with respect xi
            sig (array): standard deviation of variables
        Returns:
            std_g (float): standard deviation of g(Xi)
        """
        std_g = 0
        for dg_dx, s in zip(grad_g_x, sig):
            std_g += (dg_dx ** 2 * s ** 2)
        std_g = std_g ** (1 / 2)
        return std_g

    def run(self):
        iteration = 1
        convergence = False
        mean = []
        individuals = self.generate()
        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

            self.problem.data_store.sync_individual(individual)
            mean.append(np.mean(individual.vector))
        # X = self.generate()
        # initialize beta and alpha
        grad_g_x = self.derivative(self.problem, self.mean)
        self.evaluate(individuals)
        self.make_individual()
        mu_g = self.problem.evaluate(self.mean)
        grad_g_x = self.derivative(self.problem, self.mean)
        sig_g = self.std_linear(grad_g_x, self.std)
        beta = mu_g / sig_g
        alpha = - grad_g_x * self.std / sig_g

        # Initialize design points
        individuals = self.mean + beta * self.std * alpha
        start = time.time()
        self.problem.logger.info("CMA_ES: {}/{}".format(self.options['max_population_number'],
                                                        self.options['max_population_size']))
        while not convergence:

            for it in range(self.options['max_population_number']):
                # for i, dist in enumerate(X.dist_name):
                #     if dist != 'norm':
                #         # equivalent values
                #         X.std[i] = (1 / (X.dist_func[i].pdf(x[i]))
                #                     * stats.norm.pdf(
                #                     stats.norm.ppf(
                #                         X.dist_func[i].cdf(x[i]))))
                #         X.mean[i] = x[i] - X.std[i] * (stats.norm.ppf(
                #             X.dist_func[i].cdf(x[i])))

                # transform to standard space
                transform_individuals = (individuals - self.mean) / self.std

                # compute gradient of g with respect to z
                grad_g_x = self.derivative(self.problem, individuals)
                grad_g_z = grad_g_x * self.std

                beta_previous = beta
                beta = (self.problem.evaluate(*individuals) - grad_g_z @ transform_individuals) \
                       / np.linalg.norm(grad_g_z)
                alpha = - grad_g_z / np.linalg.norm(grad_g_z)

                # update design points in standard space
                transform_individuals = alpha * beta
                # transform to physical space
                physical_transform = self.mean + transform_individuals * self.std

                condition1 = np.linalg.norm(physical_transform - individuals)\
                             / np.linalg.norm(physical_transform) < self.tol
                condition2 = abs(np.round(beta, 3) - np.round(beta_previous, 3)) < self.tol
                if condition2 or condition1:
                    convergence = True
                if not convergence:
                    individuals = physical_transform
                    # iteration += 1
                    for individual in individuals:
                        # add to population
                        individual.population_id = it + 1
                        # append to problem
                        self.problem.individuals.append(individual)
                        # sync to datastore
                        self.problem.data_store.sync_individual(individual)

                # for individual in individuals:
                #     # add to population
                #     individual.population_id = it + 1
                #     # append to problem
                #     self.problem.individuals.append(individual)
                #     # sync to datastore
                #     self.problem.data_store.sync_individual(individual)

        return individuals, beta, iteration
